
#include <config.h>
#include <string.h>
#include <zlib.h>
#include <gio/gio.h>
#include <glib.h>
#include <glib/gprintf.h>

#include "i18n.h"
#include "proxy_server.h"

static void proxy_log_append(ProxyServer * server , time_t time,const gchar * http_request, gpointer user_data)
{
	char * strtime = ctime(&time);

	*strstr(strtime, "\n") = 0;

	g_printf("[%s]  %s\n",strtime ,http_request);
}

static GMainLoop * loop;

int main(int argc,char * argv[])
{
	gchar *domain_dirname = NULL;

	g_type_init();
	setlocale(LC_ALL,"");
	textdomain(GETTEXT_PACKAGE);

	GOptionEntry autoproxy_args[] =
	{
			{"locale",'\0',0,G_OPTION_ARG_STRING,&domain_dirname,N_("翻译目录"),N_("目录")},
			{0}
	};

	if(domain_dirname)
		bindtextdomain(GETTEXT_PACKAGE,domain_dirname);

	//创建 GUI
	ProxyServer * server = proxy_server_new();

	//创建监听端口
	if(!proxy_server_is_by_systemd(server))
	{
		GError * err=NULL;
		if (!proxy_server_add_port(server, 8080,&err))
		{
			g_error("%s",err->message);
		}

		if (!proxy_server_add_port(server, 1080,&err))
		{
			g_warning("%s",err->message);
		}
	}else
	{
		gboolean check_quit()
		{
			if ( g_atomic_int_exchange_and_add(&idle_time,-1) <= 0 )
			{
				g_main_loop_quit(loop);
				return FALSE;
			}
			return TRUE;
		}

		g_debug("Started by systemd on demand");

		g_timeout_add_seconds(1,(GSourceFunc)check_quit,0);
	}

	g_signal_connect(server,"log",G_CALLBACK(proxy_log_append),NULL);


	proxy_server_start(server);

	loop = g_main_loop_new(NULL,FALSE);
	g_main_loop_run(loop);
	g_main_loop_unref(loop);

	g_debug("Timed out waiting for new connectins , exit");
	return 0;
}

volatile gint idle_time = 60 ;

