/*
 * g_socket_source.c
 *
 *  Created on: 2010-5-15
 *      Author: cai
 */
#include "config.h"
#include <gio/gio.h>
#include "i18n.h"
#include "g_socket_source.h"

typedef struct timed_poll_source{
	GSource source;
	GSocket * socket;
	GPollFD pfd;
	GIOCondition condition;
	GTimer* last_call_back;
	gdouble	timeout;
}timed_poll_source;

static gboolean check(GSource    *_source)
{
	gulong msec;
	timed_poll_source * source = (timed_poll_source*)_source;

	if(g_timer_elapsed(source->last_call_back,&msec)>=source->timeout)
	{
		return TRUE;
	}

	g_poll(&source->pfd,1,0);

	GIOCondition condition = source->pfd.revents ;

	if(condition)
		return TRUE;
	return FALSE;
}

static gboolean prepare(GSource    *_source,gint       *timeout_)
{
	timed_poll_source * source = (timed_poll_source*)_source;
	gulong msec;
	*timeout_ = 1000 * (source->timeout -g_timer_elapsed(source->last_call_back,&msec));
	return FALSE; //check(_source);
}

static gboolean dispatch(GSource *_source, GSourceFunc _callback, gpointer user_data)
{
	timed_poll_source * source = (timed_poll_source*)_source;
	GSocketSourceFunc callback = (gpointer)_callback;
	g_timer_start(source->last_call_back );
	if(callback)
		return callback(source->socket,g_socket_condition_check(source->socket,source->condition), user_data);
	return FALSE;
}

static void finalize(GSource    *_source)
{
	timed_poll_source * source = (timed_poll_source*)_source;
	g_timer_destroy(source->last_call_back);
	g_object_unref(source->socket);
}

GSource *  g_socket_create_timeout_source(GSocket * socket ,  GIOCondition condition, guint timout)
{
	static GSourceFuncs funcs = {
			  . prepare = prepare,
			  . check = check,
			  . dispatch = dispatch,
			  . finalize = finalize
	};


	timed_poll_source * source = (timed_poll_source*) g_source_new(&funcs,sizeof(timed_poll_source));
	GPollFD * pfd = &source->pfd;

	pfd->fd = g_socket_get_fd(socket);
	pfd->events = condition;
	g_source_add_poll((GSource*)source,pfd);

	g_object_ref(socket);

	source->socket = socket;
	source->timeout = (gdouble)timout/1000;
	source->condition = condition;
	source->last_call_back = g_timer_new();

	return (GSource*)source;
}

guint g_socket_add_watch(GSocket* socket, GIOCondition condition , guint timeout , GSocketSourceFunc func,gpointer source_object)
{
	GSource * source ;

	if(source_object)
		g_object_ref(source_object);
#ifdef HAVE_DECL_G_SOCKET_SET_TIMEOUT
	source = g_socket_create_source(socket,condition,NULL); //二十五秒
	g_socket_set_timeout(socket,timeout*1000);
#else
	source = g_socket_create_timeout_source(socket,condition,timeout*1000); //二十五秒
#endif

	g_source_set_callback(source,(GSourceFunc)func,source_object,source_object?(GDestroyNotify)g_object_unref:NULL);
	guint tag = g_source_attach(source,0);
	g_source_unref(source);
	return tag;
}
