/*
 * cdnproxy.c
 *
 *  Created on: 2010-5-13
 *      Author: cai
 */
#include <config.h>

#include <sys/types.h>
#ifdef _WIN32
#include <windows.h>
/* Internet address.  */
typedef UINT in_addr_t;

#else
#include <arpa/inet.h>
#endif
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <gio/gio.h>

#include "i18n.h"
#include "proxy_session.h"
#include "proxy_server.h"
#include "g_socket_source.h"
#include "proxy_ubuntucdn.h"

struct SessionClass{
	GObjectClass parent_class;
	volatile GSocketAddress * ubuntu_cdn;
	guint			 count;
};

static gboolean start_splice(GSocket *socket,GIOCondition condition,Session * session)
{

	g_atomic_int_set(&idle_time,60);

	gsize s = 0;
	gchar buffer[15000];

	if(condition == G_IO_IN)
	{
		s = g_socket_receive(socket,buffer,sizeof(buffer),0,0);
		if(s>0)
			s = g_socket_send(session->browser,buffer,s,0,0);
	}
	return s > 0;
}

static void proxy_http_connection_modify(Session * session, const gchar* buffer)
{
	static char connection[] = "Connection:";

	char * con = strstr(buffer, connection);
	if (con)
	{
		char * start = strstr(con, ":");

		char * end = strstr(con, "\r\n");

		memset(start, ' ', end - start);

		memcpy(start, ": close", strlen(": close"));
	}
}

static gboolean transfer_http_response(GSocket *socket,GIOCondition condition,Session * session)
{
	gssize s;
	gchar buffer[1500];
	GSource* source;

	g_atomic_int_set(&idle_time,60);

	if(condition == G_IO_IN)
	{
		s = g_socket_receive(socket,buffer,sizeof(buffer),0,0);
		if(s>0)
		{
			proxy_http_connection_modify(session,buffer);

			g_socket_send(session->browser,buffer,s,0,0);

			g_socket_add_watch(socket,G_IO_IN,20,(GSocketSourceFunc)start_splice,session);
		}
	}
	return FALSE;
}

static gboolean splice_to_remote(GSocket *socket,GIOCondition condition,Session * session)
{
	g_atomic_int_set(&idle_time,60);

	gboolean wait_send(GSocket *socket,GIOCondition condition,Session * session)
	{
		g_atomic_int_set(&idle_time,60);

		if(condition == G_IO_OUT)
		{
			g_socket_add_watch(session->browser,G_IO_IN,50,(GSocketSourceFunc)splice_to_remote,session);
		}
		return FALSE;
	}

	static char buf[1024];

	int ret,fd = g_socket_get_fd(socket);
	if(condition == G_IO_IN)
	{
		ret = recv(fd,buf,sizeof(buf),MSG_PEEK);

		if (ret > 0)
		{
			ret = g_socket_send(session->cdn, buf, ret, 0, 0);

			if (ret > 0)
			{
				recv(fd, buf, ret, 0);
				return TRUE;
			}
			g_socket_add_watch(session->cdn, G_IO_OUT,25,
					(GSocketSourceFunc) wait_send, session);
			return FALSE;
		}
	}
	return FALSE;
}

static gboolean dest_host_connected(GSocket *socket,GIOCondition condition,Session * session)
{
	GSource* source;
	gssize ret;

	g_atomic_int_set(&idle_time,60);


	if(condition == G_IO_OUT)
	{
		if(!session->connected)
		{
			if (!g_socket_check_connect_result(socket, NULL))
			{
				return FALSE;
			}

			session->connected = TRUE;
		}
		//建立连接了，好，我们发送 HTTP 头数据

		switch (session->step)
		{
		case -2:
			g_socket_send(session->browser,"HTTP/1.0 200 Connection established\r\n\r\n",
					sizeof("HTTP/1.0 200 Connection established\r\n\r\n")-1,0,0);
			session->step = 0;
			break;
		case -3:
			g_socket_send(session->browser,"\005\000\000\001\000\000\000\000\000\000",10,0,0);
			session->step = 0;
			break;
		case -1:
		{
			char packet_size[20];
			sprintf(packet_size, "%d\r\n", (int) session->need_send_len);
			g_socket_send(socket, packet_size, strlen(packet_size),0,0);

			session->step++;
		}
			break;
		case 0:
			//HTTP 请求
			if (session->need_send_len == 0)
			{
				session->step++;
				return TRUE;
			}

			ret = g_socket_send(socket, session->need_send,session->need_send_len, 0, NULL);

			if (ret > 0)
			{
				memmove(session->need_send, session->need_send + ret,
						session->need_send_len - ret);
				session->need_send_len -= ret;
				g_debug("send data %d byte remain\n",(guint)session->need_send_len);
			}
			break;
		default:
			//可以开始 splice 了吗
			g_socket_add_watch(socket,G_IO_IN,100,(GSocketSourceFunc)(session->strip_response?transfer_http_response:start_splice),session);
			g_socket_add_watch(session->browser, G_IO_IN,100,(GSocketSourceFunc) splice_to_remote, session);
			return FALSE;
		}
		return TRUE;
	}
	return FALSE;
}

void proxy_session_start(Session * session)
{
	session->connected ++ ;
	if (session->connected ==2)
	{
		session->connected = 0;

		guint timeout = session->need_send_len ;

		if(( timeout / 4096 )<  5 )
			timeout = 5;
		if(timeout > 100)
			timeout = 100;

		g_socket_add_watch(session->cdn, G_IO_OUT,timeout,(GSocketSourceFunc) dest_host_connected, session);
		g_atomic_int_set(&idle_time,60);
	}
}


static void proxy_connect_to_ip(Session * session, const char ip[4])
{
	GInetAddress * inetaddr = g_inet_address_new_from_bytes(ip,G_SOCKET_FAMILY_IPV4);
	GSocketAddress * add = g_inet_socket_address_new(inetaddr,session->dest_port);
	g_socket_connect(session->cdn,add,0,0);
	g_object_unref(add);
	g_object_unref(inetaddr);
	proxy_session_start(session);
}

static void proxy_connect_to_host(Session * session, const char * host )
{
	void resolver_lookup_ready(GResolver * resolver,GAsyncResult * res, Session * session)
	{
		GList * list = g_resolver_lookup_by_name_finish(resolver,res,0);
		if (list)
		{
			GSocketAddress * add = g_inet_socket_address_new(
					g_list_first(list)->data, session->dest_port);
			g_socket_connect(session->cdn, add, 0, 0);
			g_object_unref(add);
			g_resolver_free_addresses(list);

			proxy_session_start(session);
		}
		g_object_unref(resolver);
		g_object_unref(session);
	}
	if (strlen(host))
	{
		GResolver * resolver = g_resolver_get_default();
		g_object_ref(session);
		g_resolver_lookup_by_name_async(resolver, host, 0,(GAsyncReadyCallback) resolver_lookup_ready, session);
	}
}

static gboolean proxy_connect_to_host_match(Session * session, const char * host )
{
	//根据 requset 的地址，决定连接到的地方
	void proxy_make_connection_ubuntu_cdn(Session * session)
	{
		while(PROXY_SESSION_GET_CLASS(session)->ubuntu_cdn==NULL)
		{
			g_main_context_iteration(NULL,TRUE);
		}
		g_socket_connect(session->cdn,(GSocketAddress*) PROXY_SESSION_GET_CLASS(session)->ubuntu_cdn, 0, NULL);
		session->is_ubuntu_cdn = TRUE;
		proxy_session_start(session);
	}

	if (g_regex_match_simple(".+\\.ubuntu\\.(org|com)\\.cn", host, G_REGEX_CASELESS,0) && (!g_regex_match_simple("wow\\.ubuntu\\.org\\.cn", host, G_REGEX_CASELESS,0)))
	{
		proxy_make_connection_ubuntu_cdn(session);
		return TRUE;
	}

	//开始进行 regular expansion

	proxy_connect_to_host(session,host);
	return FALSE;
}

static void proxy_get_hander(Session * session,const char * host, const char * url)
{
	if(session->http_get_hook)
	{
		session->http_get_hook(session);
	}
	if(session->is_ubuntu_cdn) //呵呵，当然不一样的啦
	{
		proxy_get_hander_ubuntucdn(session,host,url);
	}
	else
	{
		//去掉 http 请求中无用的host
		char * first_slash = strstr(strstr(session->http_header,host),"/");

		memset(session->http_header + 4 , ' ' , first_slash - session->http_header - 4);

		session->need_send = session->http_header ;
		session->need_send_len = session->http_header_len ;

		static char connection[] = "Proxy-Connection:";

		char * con = strstr(session->need_send, connection);

		if (con)
		{
			char * end = strstr(con, "\r\n");

			memset(con, ' ', end - con);

			memcpy(con,"Connection:   close",strlen("Connection:   close"));
		}

		session->http_header = NULL ;

		proxy_session_start(session);
		//		session->http_header

	}
}

static void proxy_post_hander(Session * session,const char * host, const char * url)
{
	if(session->is_ubuntu_cdn) //呵呵，当然不一样的啦
	{
		proxy_post_hander_ubuntucdn(session,host,url);
	}
	else
	{
		proxy_get_hander(session,host,url);
	}
}

static void proxy_connect_hander(Session * session,const char * host, const char * url)
{
	g_debug("got CONNECT");

	host = strtok((char*)url,":");
	session->dest_port = atoi(strtok(NULL,""));

	session->step = -2;
	proxy_session_start(session);

	proxy_connect_to_host(session,host);
}

static gboolean get_socks5_magic(GSocket *socket,GIOCondition condition,Session * session)
{
	gchar buffer[5];
	gssize ret;

	gboolean get_socks5_dnshost(GSocket *socket,GIOCondition condition,Session * session)
	{
		gchar buffer[session->http_header_len];

		memset(buffer,0,sizeof buffer);

		gssize len = g_socket_receive(socket,buffer,session->http_header_len,0,0);

		if(len <= 0)
		{
			return FALSE;
		}

		session->dest_port = ntohs( *(gushort*)(buffer+ session->http_header_len-2));
		buffer[session->http_header_len-2] = 0;

		proxy_connect_to_host_match(session,buffer);
		session->step = -3;
		proxy_session_start(session);

		gchar * des = g_strdup_printf("SOCKS5 to %s:%d",buffer,session->dest_port);
		g_signal_emit_by_name(session->proxy_server,"log",time(0),des);
		g_free(des);

		g_atomic_int_set(&idle_time,60);

		return FALSE;
	}

	gboolean get_socks5_iphost(GSocket *socket,GIOCondition condition,Session * session)
	{
		gchar buffer[6];
		gssize len;

		memset(buffer,0,sizeof buffer);

		if(condition==G_IO_IN)
		{
			len = g_socket_receive(socket,buffer,6,0,0);
			if(len < 6)
			{
				return FALSE;
			}

			gchar * des = g_strdup_printf("SOCKS5 to %s:%d",inet_ntoa( (struct in_addr){ *((in_addr_t *) buffer)}),session->dest_port);
			g_signal_emit_by_name(session->proxy_server,"log",time(0),des);
			g_free(des);

			session->step = -3;
			session->dest_port = ntohs(*(gushort*)(buffer + 4));
			proxy_connect_to_host_match(session,buffer);
			proxy_session_start(session);
		}
		return 0;
	}

	if(condition==G_IO_IN)
	{
		ret = recv(g_socket_get_fd(socket),buffer,5,MSG_PEEK);

		if(ret>=4 && buffer[0]==5 && buffer[1] == 1)
		{
			switch(buffer[3])
			{
			case 1: // IPv4
				g_socket_receive(socket,buffer,4,0,0);
				//继续读6个字节
				g_socket_add_watch(socket,G_IO_IN,3,(GSocketSourceFunc)get_socks5_iphost,session);
				break;
			case 3: // DNS 地址
				if (ret < 5)
					break;
				g_socket_receive(socket,buffer,5,0,0);
				session->http_header_len = buffer[4]+2;
				g_socket_add_watch(socket,G_IO_IN,3,(GSocketSourceFunc)get_socks5_dnshost,session);
				break;
			}
		}
		return FALSE;
	}
}

static char * get_host(char * buf, char *host)
{
	char ** chars = g_strsplit(buf, "\r\n", 100);
	char ** _chars = chars;

	for (; *_chars; _chars++) {
		if (sscanf(*_chars, "Host: %s", host) == 1) {
			g_strfreev(chars);
			return host;
		}
	}
	g_strfreev(chars);
	return 0;
}

static gssize recv_peek(GSocket *socket, char * buffer, gsize buffersize)
{
	return recv(g_socket_get_fd(socket),buffer , buffersize,MSG_PEEK);
}

static gboolean request_header_read(GSocket *socket,GIOCondition condition,Session * session)
{
	inline void proxy_log_request(Session * session, const char * buffer)
	{
		char * get = strstr(buffer,"\r\n");
		char saved = * get;
		*get = 0 ;
		time_t t;
		time(&t);
		g_signal_emit_by_name(session->proxy_server,"log",t,buffer);
		*get = saved;
	}

	gchar * newline1,*newline2;

	if (condition == G_IO_IN)
	{
		g_atomic_int_set(&idle_time,60);

		//读取
		gssize ret = recv_peek(socket,session->http_header+session->http_header_len,50000 - session->http_header_len);

		if (ret <= 0)
			return FALSE; //读取失败就不行

		newline1 = strstr(session->http_header, "\r\n\r\n");
		newline2 = strstr(session->http_header, "\n\n");
		//判断是否读取到 \r\n\r\n了
		if (newline1 || newline2) //有新行
		{
			//读取到行行为止！
			if(newline1)
				ret = (newline1 - session->http_header + 4 - session->http_header_len);
			else if ( newline2)
				ret = (newline2 - session->http_header + 2 - session->http_header_len);
			g_socket_receive(socket,session->http_header+session->http_header_len,ret,NULL,NULL);
			session->http_header_len += ret;

			//搞定。
			goto process;
			return FALSE;
		}else if(session->http_header_len + ret >= 40000){
			return FALSE; //读了那么多都没有 HTTP头？悲剧
		}else{//继续读取
			g_socket_receive(socket,session->http_header+session->http_header_len,ret,NULL,NULL);
			session->http_header_len += ret;
			return TRUE;
		}

		char host[80];
		gchar * buffer;
process:

		buffer = g_memdup(session->http_header,session->http_header_len);
		proxy_log_request(session, buffer);

		get_host(buffer,host);

		//现在处理 http 协议，支持 GET POST CONNECT
		char * method = strtok(buffer," \t\r\n");
		char * url = strtok(NULL," \t\r\n");
		char * httpver = strtok(NULL," \t\r\n");

		//首先依据主机来连接
		gboolean is_ubuntu = proxy_connect_to_host_match(session, host);

		//有跳转表
		static struct METHOD_ROUTE{
			gchar * method; // last handler is NULL, means , all
			void (*hander)(Session * session,const char * host, const char * url);
		}method_route[] = {
				{.method = "GET" , .hander =  proxy_get_hander },
				{.method = "POST" , .hander = proxy_post_hander },
				{.method = "CONNECT" , proxy_connect_hander },
				{.method = NULL, .hander= proxy_post_hander }
		};


		for (int i=0; i < G_N_ELEMENTS(method_route); i++)
		{
			if (method_route[i].method==0 || (strcmp(method,method_route[i].method) == 0))
			{
				method_route[i].hander(session, host, url);
				break;
			}
		}
		g_free(buffer);
	}
	return FALSE;
}

static gboolean request_header_socks5_read(GSocket *socket,GIOCondition condition,Session * session)
{
	gchar buffer[50001];

	if (condition == G_IO_IN)
	{
		memset(buffer,0,sizeof buffer);

		gssize ret = recv(g_socket_get_fd(socket),buffer, sizeof(buffer)-1,MSG_PEEK);
		if (ret <= 0)
			return FALSE;
		//首先看是不是 SOCKS5 协议请求
		if(buffer[0]==0x05)
		{
			gboolean get_socks5_username(GSocket *socket,GIOCondition condition,Session * session)
			{
				gchar buffer[200];

				if(G_IO_IN==condition)
				{
					gssize ret = g_socket_receive(socket,buffer,200,0,0);
				}
 				return 0;
			}

			g_debug("SOCKS5");

			for(guint count =  buffer[1] ; count ; count--)
			{
				//协议支持情况
				switch(buffer[1+count])
				{
				case 2: //用户名/密码 认证
					g_socket_receive(socket,buffer,ret,0,0);
					g_socket_send(socket,"\005\002",2,0,0);
					g_socket_add_watch(socket,G_IO_IN,5,(GSocketSourceFunc)get_socks5_username,session);
					return 0;
				case 0: //没认证
					g_socket_receive(socket,buffer,ret,0,0);
					g_socket_send(socket,"\005\000",2,0,0);
					g_socket_add_watch(socket,G_IO_IN,5,(GSocketSourceFunc)get_socks5_magic,session);
					return 0;
				}
			}

			g_atomic_int_set(&idle_time,60);
			return FALSE;
		}
		session->http_header_len = 0;
		session->http_header = g_malloc0(50001);
		g_socket_add_watch(socket,G_IO_IN,5,(GSocketSourceFunc)request_header_read,session);
	}
	return FALSE;
}


static void proxy_session_class_init(SessionClass * klass);
static void proxy_session_init(Session * session);

G_DEFINE_TYPE(Session,proxy_session,G_TYPE_OBJECT);

static void proxy_session_dispose(GObject * gobj)
{
	guint tag;
	Session * session = (Session*)gobj;

	g_object_unref(session->cdn);
	g_object_unref(session->browser);
	g_object_unref(session->proxy_server);

	session->proxy_server = NULL;

	session->cdn = session->browser = NULL ;

	G_OBJECT_CLASS(proxy_session_parent_class)->dispose(gobj);
}

static volatile gint instance = 0;

gint	proxy_session_get_instence()
{
	return instance;
}

static void proxy_session_finalize(GObject * gobj)
{
	Session * session = (Session*)gobj;

	g_free(session->http_header);
	g_free(session->need_send);

	if(session->zip_stream.state) //说明调用过 deflateInit
		deflateEnd(&session->zip_stream);
	session->zip_stream.state = NULL;

	g_debug("session %p finalized, %d left",gobj,--PROXY_SESSION_GET_CLASS(session)->count);

	G_OBJECT_CLASS(proxy_session_parent_class)->finalize(gobj);

	instance -- ;
}

static void proxy_session_init(Session * session)
{
	session->content_length = -1;
	//好了，开始连接到远程主机
	GSocket * cdn = g_socket_new(G_SOCKET_FAMILY_IPV4, G_SOCKET_TYPE_STREAM,
			G_SOCKET_PROTOCOL_DEFAULT, NULL);
	g_socket_set_blocking(cdn, FALSE);
	session->cdn = cdn;
	session->dest_port = 80;

	PROXY_SESSION_GET_CLASS(session)->count++;
}


static void proxy_session_class_init(SessionClass * klass)
{
	proxy_get_ubuntucdn_address(&klass->ubuntu_cdn);

	G_OBJECT_CLASS(klass)->finalize = proxy_session_finalize;
	G_OBJECT_CLASS(klass)->dispose = proxy_session_dispose ;
}

Session * proxy_session_new(GSocket* browser,ProxyServer * server)
{
	Session * session = g_object_new(PROXY_TYPE_SESSION,NULL);
	session->browser = browser;
	session->proxy_server = g_object_ref(server);

//	g_socket_set_blocking(browser,FALSE);
	g_socket_add_watch(browser,G_IO_IN,5,(GSocketSourceFunc)request_header_socks5_read,session);

	g_atomic_int_set(&idle_time,60);

	instance ++;

	return session;
}

