
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>
#include <gio/gio.h>
#include <gtk/gtk.h>
#include "i18n.h"
#include "proxy_server.h"
#include "proxy_session.h"

static void proxy_log_append(ProxyServer * server , time_t time,const gchar * http_request, gpointer user_data)
{
    GtkTextBuffer *buffer;
    GtkTextIter start,end;

	buffer = gtk_text_view_get_buffer(user_data);


	char * strtime = ctime(&time);

	*strstr(strtime, "\n") = 0;

	gchar * txt = g_strdup_printf("[%s]  %s\n",strtime ,http_request);

	gtk_text_buffer_get_bounds(buffer,&start,&end);

	gtk_text_buffer_insert(buffer,&end,txt,strlen(txt));

	gtk_text_view_scroll_to_iter(user_data,&end,0.1,1,0.5,0.5);

	g_free(txt);
}

static void active(GtkStatusIcon *status_icon,gpointer user_data)
{
	g_debug(__func__);

	gtk_widget_show_all(GTK_WIDGET(user_data));
	gtk_window_present(user_data);
}

static void popup_menu(GtkStatusIcon *status_icon, guint button,
		guint activate_time, gpointer user_data)
{
	g_debug(__func__);

}

void start_firefox(GtkLinkButton *button, gpointer user_data)
{
	const gchar *link_ =  gtk_link_button_get_uri(button);

	GAppLaunchContext * launcher = g_app_launch_context_new();

	g_app_info_launch_default_for_uri(link_,launcher,0);

	g_object_unref(launcher);
}

static GtkWidget * build_main_page()
{
	gint width,height;

	GtkWidget * button,*alignment;
	GtkWidget *	page =  gtk_hpaned_new();

	GtkBox * leftbox = GTK_BOX(gtk_vbox_new(1,1));
	GtkBox * rightbox = GTK_BOX(gtk_vbox_new(0,1));

	gtk_paned_pack1(GTK_PANED(page),GTK_WIDGET(leftbox),1,0);
	gtk_paned_pack2(GTK_PANED(page),GTK_WIDGET(rightbox),1,0);

	gtk_paned_set_position(GTK_PANED(page),300);

	alignment = gtk_alignment_new(.2,.2,.9,.9);
	gtk_container_add(GTK_CONTAINER(alignment),gtk_button_new_with_label(_("添加代理")));
	gtk_box_pack_start(rightbox,alignment,0,1,1);

	alignment = gtk_alignment_new(.2,.2,.9,.9);
	gtk_container_add(GTK_CONTAINER(alignment),gtk_button_new_with_label(_("生成 pac")));
	gtk_box_pack_start(rightbox,alignment,0,1,1);

	return page;
}

static GtkWidget * build_windows(ProxyServer * server)
{
	GtkWidget * main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect(main_window,"destroy",G_CALLBACK(gtk_main_quit),NULL);
	gtk_window_set_icon_name(GTK_WINDOW(main_window),GTK_STOCK_NETWORK);
	g_signal_connect(main_window,"delete-event",G_CALLBACK(gtk_widget_hide_on_delete),NULL);

	//创建系统托盘图标
	GtkStatusIcon * try_icon = gtk_status_icon_new_from_stock(GTK_STOCK_NETWORK);

	gtk_status_icon_set_tooltip_text(try_icon,_("代理自动换程序"));

	g_signal_connect(try_icon,"activate",G_CALLBACK(active),main_window);

//	g_signal_connect(try_icon,"popup-menu",G_CALLBACK(active),NULL);

	GtkWidget * note = gtk_notebook_new();
	gtk_container_add(GTK_CONTAINER(main_window),note);

	GtkWidget * action = gtk_button_new_with_label(_("退出"));
	GtkWidget * img = gtk_image_new_from_stock(GTK_STOCK_QUIT,GTK_ICON_SIZE_BUTTON);
	gtk_button_set_image(GTK_BUTTON(action),img);
	gtk_widget_show_all(action);
	gtk_notebook_set_action_widget(GTK_NOTEBOOK(note),action,GTK_PACK_END);
	g_signal_connect(action,"clicked",G_CALLBACK(gtk_main_quit),NULL);

	GtkWidget * text_view = gtk_text_view_new();
	gtk_text_view_set_editable(GTK_TEXT_VIEW(text_view),FALSE);
	GtkWidget * sw = gtk_scrolled_window_new(NULL,NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(sw),text_view);
	gtk_notebook_append_page(GTK_NOTEBOOK(note),sw,gtk_label_new(_("ubuntu论坛加速日志")));

	g_signal_connect(server,"log",G_CALLBACK(proxy_log_append),text_view);

	GtkTextBuffer * buffer = gtk_text_buffer_new(NULL);

	char * usage = _("首先查看我的博客搞定 pac : ");

	GtkTextIter start,end;
	gtk_text_buffer_get_bounds(buffer,&start,&end);
	gtk_text_buffer_insert(buffer,&end,usage,strlen(usage));

	GtkWidget * bt =gtk_link_button_new_with_label("http://microcai.gsalex.net/archives/2010/05/autoproxy-pac-firefox.html",_("pac 搞定 firefox 自动切换代理"));

	GtkTextChildAnchor * childanchor = gtk_text_buffer_create_child_anchor(buffer,&end);
	text_view = gtk_text_view_new_with_buffer(buffer);
	gtk_text_view_add_child_at_anchor(GTK_TEXT_VIEW(text_view),bt,childanchor);


	usage = _("\n接着运行就可以了。o(∩∩)o...哈哈");
	gtk_text_buffer_insert(buffer,&end,usage,strlen(usage));

	usage = _("\n当然，用");
	gtk_text_buffer_insert(buffer,&end,usage,strlen(usage));

	bt =gtk_link_button_new_with_label("https://autoproxy.org/","autoproxy");
	childanchor = gtk_text_buffer_create_child_anchor(buffer,&end);

	usage = _(" 是最好的啦！ 现在已经支持多代理了");
	gtk_text_buffer_insert(buffer,&end,usage,strlen(usage));

	gtk_text_view_add_child_at_anchor(GTK_TEXT_VIEW(text_view),bt,childanchor);

	gtk_text_view_set_editable(GTK_TEXT_VIEW(text_view),FALSE);
	gtk_notebook_append_page(GTK_NOTEBOOK(note),text_view,gtk_label_new(_("用法")));

	g_signal_connect(bt,"clicked",(GCallback)start_firefox,bt);

	gtk_window_resize(GTK_WINDOW(main_window),500,400);
	return main_window;
}

int main(int argc,char * argv[])
{
	gboolean showdefault = FALSE;
	gchar *domain_dirname = NULL;
	gint port = 8080;

	setlocale(LC_ALL,"");
	textdomain(GETTEXT_PACKAGE);

	GOptionEntry autoproxy_args[] =
	{
			{"locale",'\0',0,G_OPTION_ARG_STRING,&domain_dirname,N_("翻译目录"),N_("目录")},
			{"port",'p',0,G_OPTION_ARG_INT,&port,N_("端口"),N_("端口")},
			{"show",'w',0,G_OPTION_ARG_NONE,&showdefault,N_("显示窗口"),NULL},
			{0}
	};

	g_assert(gtk_init_with_args(&argc,&argv,NULL,autoproxy_args,GETTEXT_PACKAGE,NULL));

	if(domain_dirname)
		bindtextdomain(GETTEXT_PACKAGE,domain_dirname);

	//创建 GUI
	ProxyServer * server = proxy_server_new();
	GtkWidget * mainwin = build_windows(server);

	if(showdefault)
		gtk_widget_show_all(mainwin);

	//创建监听端口

	if(!proxy_server_is_by_systemd(server))
	{
		GError * err=NULL;
		if (!proxy_server_add_port(server, port,&err))
		{
			gtk_dialog_run((GtkDialog*)gtk_message_dialog_new(NULL,GTK_DIALOG_MODAL,GTK_MESSAGE_ERROR,GTK_BUTTONS_CLOSE,"%s",err->message));
			g_error("%s",err->message);
		}

		if (!proxy_server_add_port(server, 1080,&err))
		{
			GtkWidget * dialog = gtk_message_dialog_new(NULL,GTK_DIALOG_MODAL,GTK_MESSAGE_WARNING,GTK_BUTTONS_CLOSE,"%s",err->message);

			gtk_window_set_title(GTK_WINDOW(dialog),"Err bind to port 1080");

			g_warning("%s",err->message);
			g_timeout_add_seconds(10,(GSourceFunc)gtk_widget_destroy,dialog);
			gtk_dialog_run((GtkDialog*)dialog);
			gtk_widget_destroy(dialog);
		}
	}else
	{
		gboolean check_quit(GtkWindow * mainwin)
		{
			if ( g_atomic_int_exchange_and_add(&idle_time,-1) <= 0 )
			{
				gtk_main_quit();
				return FALSE;
			}
			return TRUE;
		}

		gboolean check_quit_2(GtkWindow * mainwin)
		{
			char * title = g_strdup_printf("%s started in systemd, exit in %d sec ( %d opening connections )",
						PACKAGE_STRING,idle_time,proxy_session_get_instence());
			gtk_window_set_title(mainwin,title);
			g_free(title);
			return TRUE;
		}

		g_debug("Started by systemd on demand");

		g_timeout_add(1000,(GSourceFunc)check_quit,mainwin);

		g_timeout_add(10,(GSourceFunc)check_quit_2,mainwin);

	}

	proxy_server_start(server);
	gtk_main();
	g_debug("Timed out waiting for new connectins , exit");
	return 0;
}

volatile gint idle_time = 60 ;

