/*
 * winmain.c
 *
 *  Created on: 2011-4-26
 *      Author: cai
 */
#include "config.h"
#include <windows.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <glib-object.h>
#include <libintl.h>

#include "proxy_server.h"
gint	proxy_session_get_instence();

static GMainLoop * loop;

static HWND hMainWin;
static GList *loglist = NULL;
static CRITICAL_SECTION cs;

static DWORD APIENTRY  win32gui_thread(PVOID param);
static void proxy_log_append(ProxyServer * server , time_t time,const gchar * http_request, gpointer user_data);


int APIENTRY WinMain(HINSTANCE hInstance,HINSTANCE notuse,LPSTR cmdline,int winmode)
{
	InitializeCriticalSection(&cs);

	CloseHandle(CreateThread(0,0,win32gui_thread,hInstance,0,0));

	gchar *domain_dirname = NULL;

	g_type_init();
	setlocale(LC_ALL,"");
	textdomain(GETTEXT_PACKAGE);

	GOptionEntry autoproxy_args[] =
	{
			{"locale",'\0',0,G_OPTION_ARG_STRING,&domain_dirname,N_("翻译目录"),N_("目录")},
			{0}
	};

	if(domain_dirname)
		bindtextdomain(GETTEXT_PACKAGE,domain_dirname);

	//创建 GUI
	ProxyServer * server = proxy_server_new();

	//创建监听端口
	GError * err=NULL;
	if (!proxy_server_add_port(server, 8080,&err))
	{
		g_error("%s",err->message);
	}

	g_signal_connect(server,"log",G_CALLBACK(proxy_log_append),NULL);

	proxy_server_start(server);

	loop = g_main_loop_new(g_main_context_default(),FALSE);
	g_main_loop_run(loop);
	g_main_loop_unref(loop);

	g_debug("Timed out waiting for new connectins , exit");
	return 0;

}

static void __stdcall do_paint(HWND hWnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	HDC	hdc;
	PAINTSTRUCT ps;
	RECT	clientRect;
	LPSTR	text = "I'm wait for connections";
	int y=0;

	GList	* itr,*Loglist = g_list_first(loglist);

	hdc = BeginPaint(hWnd,&ps);

	GetClientRect(hWnd,&clientRect);

	TextOutA(hdc,0,y,text,lstrlenA(text));

	//根据高度算应该显示多少行

	int lines = (clientRect.bottom - clientRect.top) / 20 + 1;

	if( g_list_length(Loglist) > lines)
	{
		Loglist = g_list_nth(loglist, g_list_length(loglist) - lines );
	}

	for (itr = Loglist; itr; itr = g_list_next(itr), y+= 20)
	{
		TextOutA(hdc, 0, y, itr->data, lstrlenA(itr->data));
	}

	EndPaint(hWnd,&ps);
}

static LRESULT CALLBACK WndProc(HWND hWnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	TCHAR text[200];
	switch (msg)
	{
	case WM_PAINT:
		do_paint(hWnd,msg,wparam,lparam);
		break;
	case WM_CLOSE:
		PostQuitMessage(WM_QUIT);
		break;
	case WM_TIMER:
		wsprintf(text,"%s (%d) connections",PACKAGE_STRING,proxy_session_get_instence());
		SetWindowText(hWnd,text);
	case WM_CREATE:
		SetTimer(hWnd,10,10,NULL);
	default:
		return DefWindowProc(hWnd,msg,wparam,lparam);
	}
	return TRUE;
}

void proxy_log_append(ProxyServer * server , time_t time,const gchar * http_request, gpointer user_data)
{
	char * strtime = ctime(&time);

	*strstr(strtime, "\n") = 0;

	gchar * txt = g_strdup_printf("[%s]  %s",strtime ,http_request);

	EnterCriticalSection(&cs);

	loglist = g_list_append(loglist,txt);

	InvalidateRect(hMainWin,NULL,TRUE);

	LeaveCriticalSection(&cs);

	g_debug("proxy_log_append called");

}

DWORD APIENTRY  win32gui_thread(PVOID param)
{
	HINSTANCE hInstance;

	MSG msg;
	WNDCLASS wc={0};

	LOGBRUSH brsh={0};

	hInstance = param;

	brsh.lbColor = RGB(20,50,20);

	wc.hInstance = hInstance;
	wc.lpszClassName = TEXT("autoproxy");
	wc.lpfnWndProc = WndProc;
	wc.hbrBackground = CreateBrushIndirect(&brsh);
	wc.hIcon = LoadCursor(NULL,IDC_HAND);
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.style =  CS_VREDRAW|CS_HREDRAW;

	RegisterClass(&wc);


	hMainWin = CreateWindow("autoproxy","autoproxy",WS_OVERLAPPEDWINDOW|WS_CAPTION,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,0,0,hInstance,0);

	ShowWindow(hMainWin,SW_SHOW);

	UpdateWindow(hMainWin);

	while( GetMessage(&msg,0,0,0) )
	{
		TranslateMessage(&msg);
		DispatchMessageA(&msg);
	}
	ExitProcess(0);
}

volatile int idle_time = 60 ;
